#! /bin/bash
#	Copyright (C) 2024 Canonical, Ltd.
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License as
#	published by the Free Software Foundation, version 2 of the
#	License.

#=NAME overlayfs_fuse
#=DESCRIPTION
# Verifies that file rules work in a (fuse) overlayfs
#=END

source "./overlayfs_common.inc" fuse